/* This is machine problem 1, part 2, force evaluation
 *
 * The problem is to take two sets of charged particles, 
 * where each particle has a position and a charge associated with itself,
 * and calculate the force between specific pairs of particles. 
 * An index array holds the information which particle in set B should be
 * paired with which particle in set A.
 * SUBMISSION GUIDELINES:
 * You should submit two files, called mp1-part2-solution-kernel.cu and mp1-part2-solution-host.cu
 * which contain your version of the force_eval and host_charged_particles functions.
 */

#include "../../cuda_VS_header.h"
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>

#include "../mp1_util.h"
#define EPSILON 0.00001f

// amount of floating point numbers between answer and computed value 
// for the answer to be taken correctly. 2's complement magick.
const int maxUlps = 1000;

event_pair timer;
  
float4 host_force_calc(float4 A, float4 B) {
  float x = B.x - A.x;
  float y = B.y - A.y;
  float z = B.z - A.z;
  float rsq = x*x + y*y + z*z;
  // avoid divide by zero
  if(rsq < EPSILON) {
    rsq += EPSILON;
  }
  float r = sqrt(rsq);
  float f = A.w * B.w / rsq;
  float inv_r = 1.0f / r;
  float4 fv = make_float4(x*inv_r,y*inv_r,z*inv_r,f);
  return fv;
}

//stejn funkce jako v hostovi, ale definovna pro zazen
__device__ float4 device_force_calc(float4 A, float4 B) {
  float x = B.x - A.x;
  float y = B.y - A.y;
  float z = B.z - A.z;
  float rsq = x*x + y*y + z*z;
  // avoid divide by zero
  if(rsq < EPSILON) {
    rsq += EPSILON;
  }
  float r = sqrt(rsq);
  float f = A.w * B.w / rsq;
  float inv_r = 1.0f / r;
  float4 fv = make_float4(x*inv_r, y*inv_r,z*inv_r,f);
  return fv;
}
 
void host_force_eval(float4 *set_A, float4 *set_B, int * indices, float4 *force_vectors, int array_length) {
  for(int i=0;i<array_length;i++) {
    if(indices[i] < array_length && indices[i] >= 0) {
      force_vectors[i] = host_force_calc(set_A[i], set_B[indices[i]]);
    } else {
      force_vectors[i] = make_float4(0.0, 0.0, 0.0, 0.0);
    }
  }
}

//definice kernelu na zazen
__global__ void device_force_eval(float4 *set_A, float4 *set_B, int *indices, float4 *force_vectors, int array_length) {
	int index = blockDim.x * blockIdx.x + threadIdx.x;
	if(indices[index] < array_length && indices[index] >= 0){
		force_vectors[index] = device_force_calc(set_A[index], set_B[indices[index]]);
	} else {
		force_vectors[index] = make_float4(0.0, 0.0, 0.0, 0.0);
	}
}


void host_charged_particles(float4 *h_set_A, float4 *h_set_B, int *h_indices, float4 *h_force_vectors, int num_elements) { 
	// definice pol pro zazen
	float4 *d_set_A = 0;
	float4 *d_set_B = 0;
	int *d_indices = 0;
	float4 *d_force_vectors = 0;
	// alokace pol na zazen
	int num_bytes_float = num_elements * sizeof(float4);
	int num_bytes_int = num_elements * sizeof(int);
	cudaMalloc((void**)&d_set_A, num_bytes_float);
	cudaMalloc((void**)&d_set_B, num_bytes_float);
	cudaMalloc((void**)&d_indices, num_bytes_int);
	cudaMalloc((void**)&d_force_vectors, num_bytes_float);

	// test zda se alokace zadaila
	if(d_set_A == 0 || d_set_B == 0 || d_indices == 0 || d_force_vectors == 0){
		printf("[error] Couldn't allocate CUDA memory.\n");
		exit(1);
	}

	// koprovn pol do zazen
	start_timer(&timer);
	cudaMemcpy(d_set_A, h_set_A, num_bytes_float, cudaMemcpyHostToDevice);
	cudaMemcpy(d_set_B, h_set_B, num_bytes_float, cudaMemcpyHostToDevice);
	cudaMemcpy(d_indices, h_indices, num_bytes_int, cudaMemcpyHostToDevice);
	check_launch("copy to GPU");
	stop_timer(&timer, "Copy to GPU");

	// volba potu vlken na blok
	int block_size = 128;
	int grid_size = num_elements / block_size;

	// sputn vpotu na GPU
  start_timer(&timer);
  // launch kernel
	device_force_eval<<<grid_size, block_size>>>(d_set_A, d_set_B, d_indices, d_force_vectors, num_elements);
  check_launch("gpu force eval");
  stop_timer(&timer,"Gpu force eval");

	// zskn vsledku
	start_timer(&timer);
	cudaMemcpy(h_force_vectors, d_force_vectors, num_bytes_float, cudaMemcpyDeviceToHost);
	check_launch("copy from GPU");
	stop_timer(&timer, "Copy from GPU");
  
  // uvolnn alokovanch pol
	cudaFree(d_set_A);
	cudaFree(d_set_B);
	cudaFree(d_indices);
	cudaFree(d_force_vectors);
}


int main(void) {
  // create arrays of 4M elements
  int num_elements =  1 << 22;

  // pointers to host & device arrays
  float4 *h_set_A = 0;
  float4 *h_set_B = 0;
  int *h_indices = 0;
  float4 *h_force_vectors = 0;
  float4 *h_force_vectors_checker = 0;
  
   // initialize
  srand(time(NULL)); 
  
  // malloc host array
  h_set_A = (float4*)malloc(num_elements * sizeof(float4));
  h_set_B = (float4*)malloc(num_elements * sizeof(float4));
  h_indices = (int*)malloc(num_elements * sizeof(int));
  h_force_vectors = (float4*)malloc(num_elements * sizeof(float4));
  h_force_vectors_checker = (float4*)malloc(num_elements * sizeof(float4));
  
  // if either memory allocation failed, report an error message
  if(h_set_A == 0 || h_set_B == 0 || h_force_vectors == 0 || h_indices == 0 || h_force_vectors_checker == 0) {
    printf("couldn't allocate memory\n");
    exit(1);
  }

  // generate random input
  for(int i=0;i< num_elements;i++) {
    h_set_A[i] = make_float4(rand(),rand(),rand(),rand()); 
    h_set_B[i] = make_float4(rand(),rand(),rand(),rand());

    // some indices will be invalid
    h_indices[i] = rand() % (num_elements + 2);
  }
  
  start_timer(&timer);
  // generate reference output
  host_force_eval(h_set_A, h_set_B, h_indices, h_force_vectors_checker, num_elements);
  
  check_launch("host force eval");
  stop_timer(&timer,"Host force eval");
  
  // the results of the calculation need to end up in h_force_vectors;
  host_charged_particles(h_set_A, h_set_B, h_indices, h_force_vectors, num_elements);
  
  // check CUDA output versus reference output
  int error = 0;
  
  for(int i=0;i<num_elements;i++) {
    float4 v = h_force_vectors[i];
    float4 vc = h_force_vectors_checker[i];

    if( !AlmostEqual2sComplement(v.x,vc.x,maxUlps) ||
    	!AlmostEqual2sComplement(v.y,vc.y,maxUlps) ||
    	!AlmostEqual2sComplement(v.z,vc.z,maxUlps) ||
    	!AlmostEqual2sComplement(v.w,vc.w,maxUlps)) { 
      error = 1;
    }
  }
  printf("\n");
  
  if(error) {
    printf("Output of CUDA version and normal version didn't match! \n");
  } else {
    printf("Worked! CUDA and reference output match. \n");
  }
 
  // deallocate memory
  free(h_set_A);
  free(h_set_B);
  free(h_indices);
  free(h_force_vectors);
  free(h_force_vectors_checker);
}
